﻿//Contains the information of the syntax.
//Basically, the text will be parsed into groups
//and this file defines each group, subgroup, etc.
//for easier compiling, so we need to break
//everything down into every possible group

namespace zcompiler {
	class AST {

        public abstract class Stmt {

			//Put anything necessary for statements here

		}

		public abstract class Type : object /* Inheritance needed? */ {
		
			//Put anything necessary for types here

		}

		public abstract class Expr { 
		
			//Put anything necessary for expressions here
		
		}

        public abstract class Loop {

            //Put anything necessary for loops here

        }

        public abstract class Class { 
        
            //Put anything necessary for classes here
        
        }

        public enum DoType {

            While = 0,
            Until = 1,

        }

        public enum BinaryOp {

            Add = 0,
            Sub = 1,
            Mult = 2,
            Div = 3,
            Power = 4,
            Mod = 5,

        }

        /*** --- Operations --- ***/

		//[type] [name] = [expr]
		public class Define : Stmt {

			public Type Type;
			public string Name;
			public Expr Expr;

		}

        //do [[ {{ code }} ]] <while/until> ([condition]);
        public class DoLoop : Stmt {

            public DoType Type;
            public Stmt Condition;
            public Stmt[] Code;

        }

        //for ([begin],[end],[increment]) [[ {{ code }} ]]
        public class ForLoop : Stmt {

            public int Begin;
            public int End;
            public int Increment;
            public Stmt[] Code;

        }

        //foreach ([type] [var] in [collection]) [[ {{ code }} ]]
        public class ForEachLoop : Stmt {

            public Type VarType;
            public object Var;          //Should this be an object?
            public object[] Collection;
            public Stmt[] Code;

        }

        //if ([condition]) [[ {{ code }} ]]
        public class Conditional : Stmt {

            public Stmt Condition;
            public Stmt[] Code;
            //Idea for elseif's and else's:
            //some kind of collection?
        }

        //Method: same as a function, it just doesn't return anything
        //[modifiers] sub [name] ([arguments]) [[ {{ code }} ]]
        public class Method : Stmt {

            public Stmt[] Modifiers;
            public string Name;
            public Stmt[] Args;
            public Stmt[] Code;
        
        }

        //Function
        //[modifiers] [return type] [name] ([arguments]) [[ {{ code }} ]]
        public class Function : Stmt {

            public Stmt[] Modifiers;
            public Type Return;
            public string Name;
            public Stmt[] Args;
            public Stmt[] Code;

        }

        //[modifiers] enum [name] [[ {{ code }} ]]
        public class Enum : Stmt {

            public Stmt[] Modifiers;
            public string Name;
            public Stmt[] Code;
        
        }

        //[var] = [var/num] + [var/num]
        public class Operation : Expr {

            public Expr Target;
            public Expr Left;   //*|Meaning the result of some kind of expr.
            public Expr Right;  //*|If it's something like ++ or +=, then
                                //*|"Left" = "Target"
        }

        /*** --- Data types --- ***/

        //<variable> = <expr>
        public class Variable : Expr {

            public string Ident;

        }

        //str
        public class String : Expr {

            public string Value;
        
        }

        //int
        public class Integer : Expr {

            public int Value;
        
        }

        //bool
        public class Boolean : Expr {

            public bool Value;
        
        }

        //byte
        public class Byte : Expr {

            public byte Value;
        
        }
	}
}
