﻿
//Contains the information of the syntax
//AST = Abstract Syntax Tree

namespace zcompiler {
	public class AST {

		public abstract class Stmt {

			//Put anything necessary for statements here

		}

		public abstract class Expr { 
		
			//Put anything necessary for expressions here
		
		}

		public abstract class Loop {

			//Put anything necessary for loops here

		}

		public enum DoType {

			While = 0,
			Until = 1,

		}

		public enum BinaryOp {

			Add = 0,
			Sub = 1,
			Mult = 2,
			Div = 3,
			Power = 4,
			Mod = 5,

		}

        /*** --- Constants --- ***/

        public class Constant : Stmt {

            private char val;
            public char Value {
                get { return this.val; }
            }

            public Constant(char value) {
                this.val = value;
            }
        
        }

        public static Constant Add = new Constant('+');             //Addition
        public static Constant Sub = new Constant('-');             //Subtraction
        public static Constant Mult = new Constant('*');            //Multiplication
        public static Constant Div = new Constant('/');             //Division
        public static Constant Power = new Constant('^');           //Power
        public static Constant Mod = new Constant('%');             //Modulus
        public static Constant Equal = new Constant('=');           //Equals
        public static Constant Smcln = new Constant(';');           //Semi-colon
        public static Constant Cln = new Constant(':');             //Colon (accessor)
        public static Constant ArgStart = new Constant('(');        //Argument starter
        public static Constant ArgEnd = new Constant(')');          //Argument ender
        public static Constant CBStart = new Constant('{');         //Code block starter
        public static Constant CBEnd = new Constant('}');           //Code block ender
        public static Constant VarAccStart = new Constant('[');     //Variable access starter
        public static Constant VarAccEnd = new Constant(']');       //Variable access ender
        public static Constant Dmnsn = new Constant('#');           //Dimension for arrays
        public static Constant Concat = new Constant('&');          //Concatenate for strings

		/*** --- Modifiers --- ***/

		public class Modifier : Stmt {

			public Modifiers Type;
		
		}

		public enum Modifiers { 
		
			Internal = 0,
			Global = 1,
			Closed = 2,
			Open = 3,
		
		}

		/*** --- Types --- ***/

        public enum Type { 
        
            String = 0,
            Int = 1,
            Bool = 2,
            Byte = 3,
        
        }

        //[modifiers] class [name]([args])
        public class Class : Stmt {

            public Modifiers[] Modifiers;
            public string Name;

            public Modifiers GetModifier(string mod) {

                if (mod.Equals("global"))
                    return AST.Modifiers.Global;
                else if (mod.Equals("internal"))
                    return AST.Modifiers.Internal;
                else if (mod.Equals("closed"))
                    return AST.Modifiers.Closed;
                else if (mod.Equals("open"))
                    return AST.Modifiers.Open;
                else
                    throw new System.Exception(string.Format(
                        "{0} is an unidentifiable modifier.",
                        mod));
            
            }
        
        }

		/*** --- Operations --- ***/

		//[type] [name] = [expr]
		public class Define : Stmt {

			public Type Type;
			public string Name;
			public Expr Expr;

		}

		//do [[ {{ code }} ]] <while/until> ([condition]);
		public class DoLoop : Stmt {

			public DoType Type;
			public Stmt Condition;

		}

		//for ([begin],[end],[increment]) [[ {{ code }} ]]
		public class ForLoop : Stmt {

			public int Begin;
			public int End;
			public int Increment;

		}

		//foreach ([type] [var] in [collection]) [[ {{ code }} ]]
		public class ForEachLoop : Stmt {

			public Type VarType;
			public object Var;          //Should this be an object?
			public object[] Collection;

		}

		//if ([condition]) [[ {{ code }} ]]
		public class Conditional : Stmt {

			public Stmt[] Condition;
			//Idea for elseif's and else's:
			//some kind of collection?
		}

		//Method: same as a function, it just doesn't return anything
		//[modifiers] sub [name] ([arguments]) [[ {{ code }} ]]
		public class Method : Stmt {

			public Stmt[] Modifiers;
			public string Name;
			public Stmt[] Args;
		
		}

		//Function
		//[modifiers] [return type] [name] ([arguments]) [[ {{ code }} ]]
		public class Function : Stmt {

			public Stmt[] Modifiers;
			public Type Return;
			public string Name;
			public Stmt[] Args;

		}

		//[modifiers] enum [name] [[ {{ code }} ]]
		public class Enum : Stmt {

			public Stmt[] Modifiers;
			public string Name;
		
		}

		//[var] = [var/num] + [var/num]
		public class Operation : Expr {

			public Expr Target;
			public Expr Left;   //*|Meaning the result of some kind of expr.
			public Expr Right;  //*|If it's something like ++ or +=, then
								//*|"Left" = "Target"
		}

		/*** --- Data types --- ***/

		//<variable> = <expr>
		public class Identity : Expr {

			public string Ident;

		}

		//str
		public class String : Expr {

			public string Value;
		
		}

		//int
		public class Integer : Expr {

			public int Value;
		
		}

		//bool
		public class Boolean : Expr {

			public bool Value;
		
		}

		//byte
		public class Byte : Expr {

			public byte Value;
		
		}
	}
}
