﻿using System;
using System.Collections.Generic;
using System.Text;
using zcompiler;

//Parses tokens into usable code for codegen

namespace zcompiler {
    public class Parser {

        private int index;
        private IList<object> tokens;

        private readonly List<AST.Stmt> result;
        public List<AST.Stmt> Result {
            get { return result; }
        }

        //Default constructor
        public Parser(IList<object> Tokens) {

            //Set our values
            this.tokens = Tokens;
            this.index = 0;
            this.result = new List<AST.Stmt>();

            //For every token available (index is modified by method)
            while (this.index <= this.tokens.Count) {

                //Add it to the result
                this.result.Add(this.ParseStmt());
                Console.WriteLine(String.Format("Indices up to {0} successfully parsed.", new object[] { this.index - 1 }));
            
            }

            //Make sure we've reached the EOF
            if (this.index != (this.tokens.Count) + 1)
                throw new Exception("EOF expected.");
        }

        private AST.Stmt ParseStmt() {

            //Setup our result
            AST.Stmt result;

            //If we're at the end of the data, we can't go further
            if (this.index == this.tokens.Count)
                throw new Exception("EOF found, however, an expression is expected.");

            /*** +++ Begin modifier checking +++ ***/
            #region Modifier checking
            
            //If the token is the 'internal' modifier
            if (this.tokens[this.index].Equals("internal")) {

                //Return a modifier token
                this.index++;
                AST.Modifier mod = new AST.Modifier();
                mod.Type = AST.Modifiers.Internal;
                result = mod;

            }

            //If the token is the 'global' modifier
            else if (this.tokens[this.index].Equals("global")) {

                //Return a global token
                this.index++;
                AST.Modifier mod = new AST.Modifier();
                mod.Type = AST.Modifiers.Global;
                result = mod;

            }

            //If the token is the 'internal' modifier
            else if (this.tokens[this.index].Equals("closed")) {

                //Return a modifier token
                this.index++;
                AST.Modifier mod = new AST.Modifier();
                mod.Type = AST.Modifiers.Closed;
                result = mod;

            }

            //If the token is the 'global' modifier
            else if (this.tokens[this.index].Equals("open")) {

                //Return a 'global' token
                this.index++;
                AST.Modifier mod = new AST.Modifier();
                mod.Type = AST.Modifiers.Open;
                result = mod;

            }
            #endregion
            /*** --- End modifier checking --- ***/

            /*** +++ Begin variable declaration checking +++ ***/
            #region Variable declaration checking

            //If the user is defining a string
            else if (this.tokens[this.index].Equals("str"))
            {

                //Increment to var name
                this.index++;

                //Setup a new string
                AST.Define def = new AST.Define();
                def.Type = AST.Type.String;

                //Get the variable name
                if (this.index < this.tokens.Count &&
                    this.tokens[this.index] is string)
                    def.Name = (string)this.tokens[this.index];
                else
                    throw new Exception("Expected variable name after string declaration.");

                //Increment to equals sign
                this.index++;
                
                //Check if we're still in the file
                if (this.index == this.tokens.Count)
                    throw new Exception(String.Format(
                        "End of file found, however no value for string '{0}' found.",
                        def.Name
                        ));

                //Check if the assignment character is found
                if (this.tokens[this.index] is char)
                    if ((char)this.tokens[this.index] != AST.Equal.Value)
                        throw new Exception(String.Format(
                            "Expected '{0}' after string declaration.",
                            AST.Equal.Value
                            ));

                //Increment to expression
                this.index++;
                //Get the expression
                def.Expr = this.ParseExpr();

                result = def;
            }

            //If the user is defining an int
            else if (this.tokens[this.index].Equals("int"))
            {

                //Increment to var name
                this.index++;

                //Setup a new string
                AST.Define def = new AST.Define();
                def.Type = AST.Type.Int;

                //Get the variable name
                if (this.index < this.tokens.Count &&
                    this.tokens[this.index] is int)
                    def.Name = (string)this.tokens[this.index];
                else
                    throw new Exception("Expected variable name after integer declaration.");

                //Increment to equals sign
                this.index++;

                //Check if we're still in the file
                if (this.index == this.tokens.Count)
                    throw new Exception(String.Format(
                        "End of file found, however no value for integer '{0}' found.",
                        def.Name
                        ));

                //Check if the assignment character is found
                if (this.tokens[this.index] is char)
                    if ((char)this.tokens[this.index] != AST.Equal.Value)
                        throw new Exception(String.Format(
                            "Expected '{0}' after integer declaration.",
                            AST.Equal.Value
                            ));

                //Increment to expression
                this.index++;
                //Get the expression
                def.Expr = this.ParseExpr();

                result = def;
            }

            //If the user is defining a bool
            else if (this.tokens[this.index].Equals("bool"))
            {

                //Increment to var name
                this.index++;

                //Setup a new string
                AST.Define def = new AST.Define();
                def.Type = AST.Type.Bool;

                //Get the variable name
                if (this.index < this.tokens.Count &&
                    this.tokens[this.index] is bool)
                    def.Name = (string)this.tokens[this.index];
                else
                    throw new Exception("Expected variable name after Boolean declaration.");

                //Increment to equals sign
                this.index++;

                //Check if we're still in the file
                if (this.index == this.tokens.Count)
                    throw new Exception(String.Format(
                        "End of file found, however no value for Boolean '{0}' found.",
                        def.Name
                        ));

                //Check if the assignment character is found
                if (this.tokens[this.index] is char)
                    if ((char)this.tokens[this.index] != AST.Equal.Value)
                        throw new Exception(String.Format(
                            "Expected '{0}' after Boolean declaration.",
                            AST.Equal.Value
                            ));

                //Increment to expression
                this.index++;
                //Get the expression
                def.Expr = this.ParseExpr();

                result = def;
            }

            //If the user is defining a byte
            else if (this.tokens[this.index].Equals("byte"))
            {

                //Increment to var name
                this.index++;

                //Setup a new string
                AST.Define def = new AST.Define();
                def.Type = AST.Type.Byte;

                //Get the variable name
                if (this.index < this.tokens.Count &&
                    this.tokens[this.index] is string)
                    def.Name = (string)this.tokens[this.index];
                else
                    throw new Exception("Expected variable name after byte declaration.");

                //Increment to equals sign
                this.index++;

                //Check if we're still in the file
                if (this.index == this.tokens.Count)
                    throw new Exception(String.Format(
                        "End of file found, however no value for byte '{0}' found.",
                        def.Name
                        ));

                //Check if the assignment character is found
                if (this.tokens[this.index] is char)
                    if ((char)this.tokens[this.index] != AST.Equal.Value)
                        throw new Exception(String.Format(
                            "Expected '{0}' after byte declaration.",
                            AST.Equal.Value
                            ));

                //Increment to expression
                this.index++;
                //Get the expression
                def.Expr = this.ParseExpr();

                result = def;
            }
            #endregion
            /*** --- End variable declaration checking --- ***/

            /*** +++ Begin loop checking +++ ***/
            #region Loop checking

            #endregion
            /*** --- End loop checking --- ***/

            /*** +++ Begin other keywords +++ ***/
            #region Other keyword checking

            //Check for class declaration
            else if (this.tokens[this.index].Equals("class"))
            {

                //Setup the class
                AST.Class cls = new AST.Class();

                //Get the modifiers (can be 2 at max)
                AST.Modifiers[] mods = new AST.Modifiers[2] { 0, 0 };
                int i = this.index - 1;
                int aI = 0;

                //Until we get to the "code ender" token ('}')
                //Though the cast may throw an error, we can expand on it (the error) later
                while ((string)this.tokens[this.index - i] != "}")
                {

                    //If we're in the token list
                    if (i >= 0 &&
                        this.tokens[this.index - i] is string &&
                        aI <= mods.Length)
                        //Get the modifier
                        mods[aI] = cls.GetModifier((string)this.tokens[i]);
                    else if (aI > mods.Length)
                        throw new Exception("A class may not have more than 2 modifiers.");
                    else
                        //Tell user we cannot find any modifiers
                        throw new Exception(String.Format(
                            "Cannot find any proper definitions for class {0}.",
                            new object[] { (string)this.tokens[this.index + 2] }));

                    i--;
                    aI++;

                    //If we're outside the tokens, quit the loop
                    if (i < 0)
                        break;

                }

                //If there are no modifiers
                if (mods.Length == 0)
                    throw new Exception(String.Format(
                        "Cannot find any proper definitions for class {0}.",
                        new object[] { (string)this.tokens[this.index + 2] }));

                //Set the modifiers
                cls.Modifiers = mods;

                //Increment to class name
                this.index++;

                //Check if we're in the file
                if (this.index < this.tokens.Count &&
                    this.tokens[this.index] is string)
                {
                    //Get the class name
                    cls.Name = (string)this.tokens[this.index];
                }
                else
                    throw new Exception("Expected name for class after class definition.");

                //Increment past the class name
                this.index++;

                //Return the class
                result = cls;

            }

            #endregion
            /*** --- End other keywords --- ***/

            /*** +++ Begin char checking +++ ***/
            #region Constant checking
            
            //First, make sure it is a constant
            else if (this.tokens[this.index] is AST.Constant) {

                //Return the value
                result = (AST.Constant)this.tokens[this.index];

                //Go to the next token
                this.index++;
            }
            
            #endregion
            /*** --- End char checking --- ***/

            //If there was an error somewhere
            else {
                throw new Exception(String.Format(
                    "Parse error; unrecognized token.\nIndex: {0}\nToken: \"{1}\"",
                    new object[] { this.index, this.tokens[this.index] }
                    ));
            }

            return result;
        }

        /// <summary>
        /// Converts current token into an expression.
        /// </summary>
        /// <returns>The current token as an expression.</returns>
        private AST.Expr ParseExpr() {

            //Check if we've reached the end of the tokens, first
            if (this.index == this.tokens.Count)
                throw new Exception("EOF has been found, however an expression is expected.");

            //If the data is a raw string, i.e. we didn't do
            //<type>.Parse() or data.ToString()
            if (this.tokens[this.index] is StringBuilder) {

                //Get that string
                string value = ((StringBuilder)this.tokens[this.index++]).ToString();
                //Create a new Zenon string
                AST.String str = new AST.String();
                //Set the Zenon string's value
                str.Value = value;
                //Return that string
                return str;

            }
            //If the data is an integer
            else if (this.tokens[this.index] is int) {

                //Get that integer
                int value = (int)this.tokens[this.index++];
                //Create a new Zenon integer
                AST.Integer anInt = new AST.Integer();
                //Set it's value
                anInt.Value = value;
                //Return it
                return anInt;

            }
            //If the data is a variable name or keyword
            else if (this.tokens[this.index] is string) {

                //Get the variable
                string identity = (string)this.tokens[this.index++];
                //Create a new Zenon var
                AST.Identity var = new AST.Identity();
                //Get its identity
                var.Ident = identity;
                //Return it
                return var;

            }
            //If the data is a byte
            else if (this.tokens[this.index] is byte) {

                //Get that byte
                byte value = (byte)this.tokens[this.index++];
                //Create a new Zenon byte
                AST.Byte zbyte = new AST.Byte();
                //Store the value
                zbyte.Value = value;
                //Return it
                return zbyte;

            }
            //If the data is a Boolean value
            else if (this.tokens[this.index] is bool) {

                //Get that bool
                bool value = (bool)this.tokens[this.index++];
                //Create a Zenon bool
                AST.Boolean zbool = new AST.Boolean();
                //Set its value
                zbool.Value = value;
                //Return it
                return zbool;
            
            }
            else
                throw new Exception("Expected string, int, byte, bool, or variable.");
        
        }

        /// <summary>
        /// Returns the statement equivalent to the provided character.
        /// </summary>
        /// <param name="chr">Character to check for statement equality.</param>
        public AST.Stmt CharConvert(char chr) {

            //Setup our return variable
            AST.Stmt result;

            //Check for possible valid values
            //Addition
            if (chr == AST.Add.Value)
                result = AST.Add;
            //Argument end
            else if (chr == AST.ArgEnd.Value)
                result = AST.ArgEnd;
            //Argument start
            else if (chr == AST.ArgStart.Value)
                result = AST.ArgStart;
            //Code block end
            else if (chr == AST.CBEnd.Value)
                result = AST.CBEnd;
            //Code block start
            else if (chr == AST.CBStart.Value)
                result = AST.CBStart;
            //Colon (accessor)
            else if (chr == AST.Cln.Value)
                result = AST.Cln;
            //Concatenation
            else if (chr == AST.Concat.Value)
                result = AST.Concat;
            //Division
            else if (chr == AST.Div.Value)
                result = AST.Div;
            //Dimension (for arrays)
            else if (chr == AST.Dmnsn.Value)
                result = AST.Dmnsn;
            //Equals
            else if (chr == AST.Equal.Value)
                result = AST.Equal;
            //Modulus
            else if (chr == AST.Mod.Value)
                result = AST.Mod;
            //Multiplication
            else if (chr == AST.Mult.Value)
                result = AST.Mult;
            //... to the power of ...
            else if (chr == AST.Power.Value)
                result = AST.Power;
            //Semicolon (line end)
            else if (chr == AST.Smcln.Value)
                result = AST.Smcln;
            //Subtraction
            else if (chr == AST.Sub.Value)
                result = AST.Sub;
            //Variable access end
            else if (chr == AST.VarAccEnd.Value)
                result = AST.VarAccEnd;
            //Variable access start
            else if (chr == AST.VarAccStart.Value)
                result = AST.VarAccStart;
            //If we can't find the char, throw an error
            else
                throw new Exception(String.Format(
                    "{0} is an invalid constant.",
                    chr
                    ));

            //Return the result
            return result;
        }
    }
}